// ignore_for_file: public_member_api_docs

import 'dart:developer';

import 'package:fl_chart/fl_chart.dart';
import 'package:healthrian_ble_support_for_wearecg12/'
    'healthrian_ble_support_for_wearecg12.dart';
import 'package:healthrian_common_support/healthrian_common_support.dart';
import 'package:lzma/lzma.dart';

mixin Ecg {
  /// [maxBufferStreamEcg] decides a maximum length of buffered Stream.
  ///
  /// For 40 seconds buffer and 250 sampling rate,
  /// It uses 40 * 250.
  int maxBufferStreamEcg = 40 * 250;

  /// [lengthBufferStreamEcg] decides a length of buffered Stream,
  /// after flushing buffer when reaching a maximum size.
  ///
  /// For 30 seconds buffer and 250 sampling rate,
  /// It uses 30 * 250.
  int lengthBufferStreamEcg = 30 * 250;

  /// [isFilteredEcg] decides whether ecg signal is filtered.
  bool isFilteredEcg = true;

  /// [isCompressedEcg] decides whether ecg signal is compressed.
  ///
  /// If it is `true`, ecg signal is decompressed by LZMA algorithm.
  bool isCompressedEcg = false;

  final _dataFilter = DataFilter(1);
}

extension BLEEcg on BLEManager {
  double _filter(double value) => _dataFilter.applyFilter(0, value, true);

  /// [startMeasurementEcg] sends read_ecg_packet to device,
  ///  and then, it registers all ecg streams of [streams].
  Future<void> startMeasurementEcg() async {
    await _sendCommandReadEcg();

    final isRunnable = await _checkValidation();
    if(!isRunnable) return;

    streams
      ..rawStreamEcg = _getStreamEcg().broadcast
      ..decompressedStreamEcg = _getDecompressedStreamEcg().broadcast
      ..recentStreamEcg = _getRecentStreamEcg().broadcast
      ..bufferedStreamEcg = _getBufferedStreamEcg().broadcast
      ..bufferedPlotStreamEcg = _getBufferedPlotStreamEcg().broadcast;
  }

  /// [stopMeasurementEcg] sends stop_ecg_packet to device.
  Future<void> stopMeasurementEcg() async {
    final control = points[Characteristics.DEVICE_CONTROL_POINT];
    if (control == null) return;

    log('Writes a packet ${Commands.packetStopEcg}');
    await control.write(Commands.packetStopEcg);
  }

  Future<bool> _checkValidation() async {
    final confirm = points[Characteristics.DEVICE_CONTROL_CONFIRM];
    if (confirm == null) return false;

    final confirmValue = await confirm.read();
    if (confirmValue.isEmpty) return false;
    if (confirmValue[0] == 2) return false;
    return true;
  }

  Future<void> _sendCommandReadEcg() async {
    final control = points[Characteristics.DEVICE_CONTROL_POINT];
    if (control == null) return;

    log('Writes a packet ${Commands.packetReadEcg}');
    await control.write(Commands.packetReadEcg);
  }

  Stream<List<int>> _getStreamEcg() async* {
    final oneChannel = points[Characteristics.ONE_CHANNEL];
    if (oneChannel == null) return;
    // var now = DateTime.now();
    await oneChannel.setNotifyValue(true);
    await for (final res in oneChannel.value) {
      yield res;
      // final newNow = DateTime.now();
      // log(newNow.difference(now).toString());
      // now = newNow;
    }
  }

  Stream<List<int>> _getDecompressedStreamEcg() async* {
    final events = streams.rawStreamEcg;
    if (events == null) return;
    await for (final event in events) {
      yield isCompressedEcg ? lzma.decode(event) : event;
    }
  }

  Stream<List<double>> _getRecentStreamEcg() async* {
    final events = streams.decompressedStreamEcg;
    if (events == null) return;
    await for (final event in events) {
      if (event.isEmpty) continue;
      final itemCount = event.length ~/ 3;
      final res = <double>[];
      for (var i = 0; i < itemCount; i++) {
        final raw = toDouble(
          event[i * 3 + 0],
          event[i * 3 + 1],
          event[i * 3 + 2],
        );
        res.add(isFilteredEcg ? _filter(raw) : raw);
      }
      yield res;
    }
  }

  Stream<List<double>> _getBufferedStreamEcg() async* {
    final events = streams.recentStreamEcg;
    if (events == null) return;
    final res = <double>[];
    await for (final event in events) {
      res.addAll(event);
      if (res.length > maxBufferStreamEcg) {
        res.removeRange(0, res.length - lengthBufferStreamEcg);
      }
      yield res;
    }
  }

  Stream<List<FlSpot>> _getBufferedPlotStreamEcg() async* {
    final events = streams.recentStreamEcg;
    if (events == null) return;
    final res = <FlSpot>[];
    var x = 0.0;
    await for (final event in events) {
      for (final i in Iterable<int>.generate(event.length)) {
        res.add(FlSpot(x + i, event[i]));
      }
      if (res.length > maxBufferStreamEcg) {
        res.removeRange(0, res.length - lengthBufferStreamEcg);
      }
      yield res;
      x += event.length;
    }
  }
}
